from bolsonaro.models.model_raw_results import ModelRawResults
from bolsonaro.error_handling.logger_factory import LoggerFactory
from . import LOG_PATH

import time
import datetime
import numpy as np


class Trainer(object):
    """
    Class capable of fitting any model object to some prepared data then evaluate and save results through the `train` method.
    """

    def __init__(self, dataset):
        """

        :param dataset: Object with X_train, y_train, X_dev, y_dev, X_test and Y_test attributes
        """
        self._dataset = dataset
        self._logger = LoggerFactory.create(LOG_PATH, __name__)

    def init(self, model):
        if model.models_parameters.subsets_used == 'train,dev':
            self._X_forest = self._dataset.X_train
            self._y_forest = self._dataset.y_train
            self._X_omp = self._dataset.X_dev
            self._y_omp = self._dataset.y_dev
            self._logger.debug('Fitting the forest on train subset and OMP on dev subset.')
        elif model.models_parameters.subsets_used == 'train+dev,train+dev':
            self._X_forest = np.concatenate([self._dataset.X_train, self._dataset.X_dev])
            self._X_omp = self._X_forest
            self._y_forest = np.concatenate([self._dataset.y_train, self._dataset.y_dev])
            self._y_omp = self._y_forest
            self._logger.debug('Fitting both the forest and OMP on train+dev subsets.')
        elif model.models_parameters.subsets_used == 'train,train+dev':
            self._X_forest = self._dataset.X_train
            self._y_forest = self._dataset.y_train
            self._X_omp = np.concatenate([self._dataset.X_train, self._dataset.X_dev])
            self._y_omp = np.concatenate([self._dataset.y_train, self._dataset.y_dev])
        else:
            raise ValueError("Unknown specified subsets_used parameter '{}'".format(model.models_parameters.subsets_used))

    def train(self, model):
        """
        :param model: Object with
        :return:
        """
        
        self._logger.debug('Training model using train set...')
        self._begin_time = time.time()
        model.fit(
            X_forest=self._X_forest,
            y_forest=self._y_forest,
            X_omp=self._X_omp,
            y_omp=self._y_omp
        )
        self._end_time = time.time()

    def compute_results(self, model, models_dir):
        """
        :param model: Object with
        :param models_dir: Where the results will be saved
        """
        results = ModelRawResults(
            model_object=model,
            training_time=self._end_time - self._begin_time,
            datetime=datetime.datetime.now(),
            train_score=model.score(self._dataset.X_train, self._dataset.y_train),
            dev_score=model.score(self._dataset.X_dev, self._dataset.y_dev),
            test_score=model.score(self._dataset.X_test, self._dataset.y_test),
            score_metric=model.DEFAULT_SCORE_METRIC, # TODO: resolve the used metric in a proper way
            train_score_regressor=model.score_base_estimator(self._dataset.X_train, self._dataset.y_train),
            dev_score_regressor=model.score_base_estimator(self._dataset.X_dev, self._dataset.y_dev),
            test_score_regressor=model.score_base_estimator(self._dataset.X_test, self._dataset.y_test)
        )
        results.save(models_dir)
        self._logger.info("Base performance on test: {}".format(results.test_score_regressor))
        self._logger.info("Performance on test: {}".format(results.test_score))

        self._logger.info("Base performance on train: {}".format(results.train_score_regressor))
        self._logger.info("Performance on train: {}".format(results.train_score))

        self._logger.info("Base performance on dev: {}".format(results.dev_score_regressor))
        self._logger.info("Performance on dev: {}".format(results.dev_score))
